/* SPDX-License-Identifier: GPL-2.0 */
/*
* This file is subject to the terms and conditions of the GNU General Public
* License.  See the file "COPYING" in the main directory of this archive
* for more details.
*
* Copyright (C) 2003, 2004 Ralf Baechle
* Copyright (C) 2004  Maciej W. Rozycki
* Copyright (C) 2020 Loongson Technology Corporation Limited
*/
#ifndef __ASM_CPU_FEATURES_H
#define __ASM_CPU_FEATURES_H

#include <asm/cpu.h>
#include <asm/cpu-info.h>
#include <asm/mach-la64/cpu-feature-overrides.h>

#define __ase(ase)			(cpu_data[0].ases & (ase))
#define __opt(opt)			(cpu_data[0].options & (opt))

#define cpu_has_loongarch		(cpu_has_loongarch32 | cpu_has_loongarch64)
#define cpu_has_loongarch32		(cpu_data[0].isa_level & LOONGARCH_CPU_ISA_32BIT)
#define cpu_has_loongarch64		(cpu_data[0].isa_level & LOONGARCH_CPU_ISA_64BIT)

#ifndef cpu_icache_line_size
#define cpu_icache_line_size()		cpu_data[0].icache.linesz
#endif
#ifndef cpu_dcache_line_size
#define cpu_dcache_line_size()		cpu_data[0].dcache.linesz
#endif
#ifndef cpu_vcache_line_size
#define cpu_vcache_line_size()		cpu_data[0].vcache.linesz
#endif
#ifndef cpu_scache_line_size
#define cpu_scache_line_size()		cpu_data[0].scache.linesz
#endif

#ifdef CONFIG_32BIT
# ifndef cpu_has_64bits
# define cpu_has_64bits			(cpu_data[0].isa_level & LOONGARCH_CPU_ISA_64BIT)
# endif
# ifndef cpu_has_64bit_zero_reg
# define cpu_has_64bit_zero_reg		(cpu_data[0].isa_level & LOONGARCH_CPU_ISA_64BIT)
# endif
# ifndef cpu_has_64bit_gp_regs
# define cpu_has_64bit_gp_regs		0
# endif
# ifndef cpu_vmbits
# define cpu_vmbits 31
# endif
#endif

#ifdef CONFIG_64BIT
# ifndef cpu_has_64bits
# define cpu_has_64bits			1
# endif
# ifndef cpu_has_64bit_zero_reg
# define cpu_has_64bit_zero_reg		1
# endif
# ifndef cpu_has_64bit_gp_regs
# define cpu_has_64bit_gp_regs		1
# endif
# ifndef cpu_vmbits
# define cpu_vmbits cpu_data[0].vmbits
# endif
# define __NEED_VMBITS_PROBE
#endif

/*
* SMP assumption: Options of CPU 0 are a superset of all processors.
* This is true for all known LOONGARCH systems.
*/
#ifndef cpu_has_tlb
#define cpu_has_tlb		__opt(LOONGARCH_CPU_TLB)
#endif

#ifndef cpu_has_csr
#define cpu_has_csr		__opt(LOONGARCH_CPU_CSR)
#endif

#ifndef cpu_has_fpu
#define cpu_has_fpu		1
#define raw_cpu_has_fpu		1
#else
#define raw_cpu_has_fpu		cpu_has_fpu
#endif

#ifndef cpu_has_llsc
#define cpu_has_llsc		__opt(LOONGARCH_CPU_LLSC)
#endif

#ifndef cpu_has_watch
#define cpu_has_watch		__opt(LOONGARCH_CPU_WATCH)
#endif

#ifndef cpu_has_vint
#define cpu_has_vint		__opt(LOONGARCH_CPU_VINT)
#endif

#ifndef cpu_has_extioi
#define cpu_has_extioi		__opt(LOONGARCH_CPU_EXTIOI)
#endif

#ifndef cpu_has_csripi
#define cpu_has_csripi		__opt(LOONGARCH_CPU_CSRIPI)
#endif

#ifndef cpu_has_prefetch
#define cpu_has_prefetch	__opt(LOONGARCH_CPU_PREFETCH)
#endif

#ifndef cpu_has_lpa
#define cpu_has_lpa		__opt(LOONGARCH_CPU_LPA)
#endif

#ifndef cpu_has_counter
#define cpu_has_counter		__opt(LOONGARCH_CPU_GFT | LOONGARCH_CPU_LFT)
#endif

#ifndef cpu_has_pmp
#define cpu_has_pmp		__opt(LOONGARCH_CPU_PMP)
#endif

#ifndef cpu_has_perf
#define cpu_has_perf		__opt(LOONGARCH_CPU_PMP)
#endif

#ifndef cpu_has_scalefreq
#define cpu_has_scalefreq	__opt(LOONGARCH_CPU_SCALEFREQ)
#endif

#ifndef cpu_has_guestid
#define cpu_has_guestid		__opt(LOONGARCH_CPU_GUESTID)
#endif

#ifndef cpu_has_hypervisor
#define cpu_has_hypervisor	__opt(LOONGARCH_CPU_HYPERVISOR)
#endif

#ifndef cpu_has_matc_guest
#define cpu_has_matc_guest	(cpu_data[0].guest_cfg & (1 << 0))
#endif
#ifndef cpu_has_matc_root
#define cpu_has_matc_root	(cpu_data[0].guest_cfg & (1 << 1))
#endif
#ifndef cpu_has_matc_nest
#define cpu_has_matc_nest	(cpu_data[0].guest_cfg & (1 << 2))
#endif
#ifndef cpu_has_sitp
#define cpu_has_sitp		(cpu_data[0].guest_cfg & (1 << 6))
#endif
#ifndef cpu_has_titp
#define cpu_has_titp		(cpu_data[0].guest_cfg & (1 << 8))
#endif
#ifndef cpu_has_toep
#define cpu_has_toep		(cpu_data[0].guest_cfg & (1 << 10))
#endif
#ifndef cpu_has_topp
#define cpu_has_topp		(cpu_data[0].guest_cfg & (1 << 12))
#endif
#ifndef cpu_has_torup
#define cpu_has_torup		(cpu_data[0].guest_cfg & (1 << 14))
#endif
#ifndef cpu_has_gcip_all
#define cpu_has_gcip_all	(cpu_data[0].guest_cfg & (1 << 16))
#endif
#ifndef cpu_has_gcip_hit
#define cpu_has_gcip_hit	(cpu_data[0].guest_cfg & (1 << 17))
#endif
#ifndef cpu_has_gcip_secure
#define cpu_has_gcip_secure	(cpu_data[0].guest_cfg & (1 << 18))
#endif

#ifndef cpu_has_cpucfg
#define cpu_has_cpucfg		__ase(LOONGARCH_ASE_CPUCFG)
#endif

#ifndef cpu_has_lamo
#define cpu_has_lamo    	__ase(LOONGARCH_ASE_LAMO)
#endif

#ifndef cpu_has_lsx
#define cpu_has_lsx		__ase(LOONGARCH_ASE_LSX)
#endif

#ifndef cpu_has_lasx
#define cpu_has_lasx		__ase(LOONGARCH_ASE_LASX)
#endif

#ifndef cpu_has_lvz
#define cpu_has_lvz		__ase(LOONGARCH_ASE_LVZ)
#endif

#ifndef cpu_has_lbt
#define cpu_has_lbt		__ase(LOONGARCH_ASE_LBT)
#endif

#ifndef cpu_has_cam
#define cpu_has_cam		__ase(LOONGARCH_ASE_CAM)
#endif

/*
* Guest dynamic capabilities
*/
#ifndef cpu_guest_has_dyn_fpu
#define cpu_guest_has_dyn_fpu	(cpu_data[0].guest.options_dyn & LOONGARCH_CPU_FPU)
#endif
#ifndef cpu_guest_has_dyn_perf
#define cpu_guest_has_dyn_perf	(cpu_data[0].guest.options_dyn & LOONGARCH_CPU_PMP)
#endif
#ifndef cpu_guest_has_dyn_lsx
#define cpu_guest_has_dyn_lsx	(cpu_data[0].guest.ases_dyn & LOONGARCH_ASE_LSX)
#endif

#endif /* __ASM_CPU_FEATURES_H */
