/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (C) 2020 Loongson Technology Corporation Limited
 *
 * This program is free software; you can redistribute	it and/or modify it
 * under  the terms of	the GNU General	 Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#ifndef LOONGSON_CPU_H_
#define LOONGSON_CPU_H_

#include <linux/types.h>
#include <linux/bits.h>

#include <asm/loongarchregs.h>
#include <asm/cpu.h>
#include <lcsrintrin.h>

/*
 * Loongson CPU Option encodings
 */
static inline u32 read_cpucfg(u32 reg)
{
	u32 __res;

	__asm__ __volatile__(
		"parse_r __res,%0\n\t"
		"parse_r reg,%1\n\t"
		".word (0x6c00 | (reg << 5) | (__res))\n\t"
		:"=r"(__res)
		:"r"(reg)
		:
		);
	return __res;
}

enum invtlb_ops {
	/* Invalid all tlb */
	INVTLB_ALL = 0x0,
	/* Invalid current tlb */
	INVTLB_CURRENT_ALL = 0x1,
	/* Invalid all global=1 lines in current tlb */
	INVTLB_CURRENT_GTRUE = 0x2,
	/* Invalid all global=0 lines in current tlb */
	INVTLB_CURRENT_GFALSE = 0x3,
	/* Invalid global=0 and matched asid lines in current tlb */
	INVTLB_GFALSE_AND_ASID = 0x4,
	/* Invalid addr with global=0 and matched asid in current tlb */
	INVTLB_ADDR_GFALSE_AND_ASID = 0x5,
	/* Invalid addr with global=1 or matched asid in current tlb */
	INVTLB_ADDR_GTRUE_OR_ASID = 0x6,
	/* Invalid matched gid in guest tlb */
	INVGTLB_GID = 0x9,
	/* Invalid global=1, matched gid in guest tlb */
	INVGTLB_GID_GTRUE = 0xa,
	/* Invalid global=0, matched gid in guest tlb */
	INVGTLB_GID_GFALSE = 0xb,
	/* Invalid global=0, matched gid and asid in guest tlb */
	INVGTLB_GID_GFALSE_ASID = 0xc,
	/* Invalid global=0 , matched gid, asid and addr in guest tlb */
	INVGTLB_GID_GFALSE_ASID_ADDR = 0xd,
	/* Invalid global=1 , matched gid, asid and addr in guest tlb */
	INVGTLB_GID_GTRUE_ASID_ADDR = 0xe,
	/* Invalid all gid gva-->gpa guest tlb */
	INVGTLB_ALLGID_GVA_TO_GPA = 0x10,
	/* Invalid all gid gpa-->hpa tlb */
	INVTLB_ALLGID_GPA_TO_HPA = 0x11,
	/* Invalid all gid tlb, including  gva-->gpa and gpa-->hpa */
	INVTLB_ALLGID = 0x12,
	/* Invalid matched gid gva-->gpa guest tlb */
	INVGTLB_GID_GVA_TO_GPA = 0x13,
	/* Invalid matched gid gpa-->hpa tlb */
	INVTLB_GID_GPA_TO_HPA = 0x14,
	/* Invalid matched gid tlb,including gva-->gpa and gpa-->hpa */
	INVTLB_GID_ALL = 0x15,
	/* Invalid matched gid and addr gpa-->hpa tlb */
	INVTLB_GID_ADDR = 0x16,
};

/*
 *invtlb op info addr
 * (0x1 << 26) | (0x24 << 20) | (0x13 << 15) |
 * (addr << 10) | (info << 5) | op
 */
static inline void invtlb_all(u32 op, u32 info, u64 addr)
{
	__asm__ __volatile__(
		".word ((0x6498000) | (0 << 10) | (0 << 5) | %0)\n\t"
		:
		:"i"(op)
		:
		);
}

static inline void invtlb_all_atomic(u32 op, u32 info, u64 addr)
{
	invtlb_all(op, info, addr);
}

static inline void invtlb_noinfo(u32 op, u32 info, u64 addr)
{
	__asm__ __volatile__(
		"parse_r addr,%0\n\t"
		".word ((0x6498000) | (addr << 10) | (0 << 5) | %1)\n\t"
		:
		:"r"(addr), "i"(op)
		:
		);
}

static inline void invtlb_noinfo_atomic(u32 op, u32 info, u64 addr)
{
	invtlb_noinfo(op, info, addr);
}

static inline void invtlb_noaddr(u32 op, u32 info, u64 addr)
{
	__asm__ __volatile__(
		"parse_r info,%0\n\t"
		".word ((0x6498000) | (0 << 10) | (info << 5) | %1)\n\t"
		:
		:"r"(info), "i"(op)
		:
		);
}

static inline void invtlb(u32 op, u32 info, u64 addr)
{
	__asm__ __volatile__(
		"parse_r addr,%0\n\t"
		"parse_r info,%1\n\t"
		".word ((0x6498000) | (addr << 10) | (info << 5) | %2)\n\t"
		:
		:"r"(addr), "r"(info), "i"(op)
		:
		);
}

static inline void invtlb_atomic(u32 op, u32 info, u64 addr)
{
	invtlb(op, info, addr);
}

/* Bit Domains for CFG registers */
#define LOONGSON_CFG0		0x0
#define LOONGSON_CFG0_PRID 	GENMASK(31, 0)

#define LOONGSON_CFG1 		0x1
#define LOONGSON_CFG1_ISGR32	BIT(0)
#define LOONGSON_CFG1_ISGR64	BIT(1)
#define LOONGSON_CFG1_PAGING	BIT(2)
#define LOONGSON_CFG1_IOCSR	BIT(3)
#define LOONGSON_CFG1_PABITS	GENMASK(11, 4)
#define LOONGSON_CFG1_VABITS	GENMASK(19, 12)
#define LOONGSON_CFG1_UAL	BIT(20)
#define LOONGSON_CFG1_RI	BIT(21)
#define LOONGSON_CFG1_XI	BIT(22)
#define LOONGSON_CFG1_RPLV	BIT(23)
#define LOONGSON_CFG1_HUGEPG	BIT(24)
#define LOONGSON_CFG1_IOCSRBRD	BIT(25)
/* 0 for line, 1 for message */
#define LOONGSON_CFG1_INTIMPL	BIT(26)

#define LOONGSON_CFG2 		0x2
#define LOONGSON_CFG2_FP	BIT(0)
#define LOONGSON_CFG2_FPSP	BIT(1)
#define LOONGSON_CFG2_FPDP	BIT(2)
#define LOONGSON_CFG2_FPVERS	GENMASK(5, 3)
#define LOONGSON_CFG2_LSX	BIT(6)
#define LOONGSON_CFG2_LASX	BIT(7)
#define LOONGSON_CFG2_COMPLEX	BIT(8)
#define LOONGSON_CFG2_CRYPTO	BIT(9)
#define LOONGSON_CFG2_LVZP	BIT(10)
#define LOONGSON_CFG2_LVZVER	GENMASK(13, 11)
#define LOONGSON_CFG2_LLFTP	BIT(14)
#define LOONGSON_CFG2_LLFTPREV	GENMASK(17, 15)
#define LOONGSON_CFG2_X86BT	BIT(18)
#define LOONGSON_CFG2_ARMBT	BIT(19)
#define LOONGSON_CFG2_MIPSBT	BIT(20)
#define LOONGSON_CFG2_LSPW	BIT(21)
#define LOONGSON_CFG2_LAMO	BIT(22)

#define LOONGSON_CFG3		0x3
#define LOONGSON_CFG3_CCDMA	BIT(0)
#define LOONGSON_CFG3_SFB	BIT(1)
#define LOONGSON_CFG3_UCACC	BIT(2)
#define LOONGSON_CFG3_LLEXC	BIT(3)
#define LOONGSON_CFG3_SCDLY	BIT(4)
#define LOONGSON_CFG3_LLDBAR	BIT(5)
#define LOONGSON_CFG3_ITLBT	BIT(6)
#define LOONGSON_CFG3_ICACHET	BIT(7)
#define LOONGSON_CFG3_SPW_LVL	GENMASK(10, 8)
#define CFG3_SPW_LVL_DEFAULT	4
#define LOONGSON_CFG3_SPW_HG_HF	BIT(11)
#define LOONGSON_CFG3_RVA	BIT(12)
#define LOONGSON_CFG3_RVAMAX	GENMASK(16, 13)

#define LOONGSON_CFG4 		0x4
#define LOONGSON_CFG4_CCFREQ	GENMASK(31, 0)

#define LOONGSON_CFG5 		0x5
#define LOONGSON_CFG5_CCMUL	GENMASK(15, 0)
#define LOONGSON_CFG5_CCDIV	GENMASK(31, 16)

#define LOONGSON_CFG6 		0x6
#define LOONGSON_CFG6_PMP	BIT(0)
#define LOONGSON_CFG6_PAMVER	GENMASK(3, 1)
#define LOONGSON_CFG6_PMNUM	GENMASK(7, 4)
#define LOONGSON_CFG6_PMBITS	GENMASK(13, 8)
#define LOONGSON_CFG6_UPM	BIT(14)

#define LOONGSON_CFGX 		0x10
#define LOONGSON_CFGX_L1_IUPRE	BIT(0)
#define LOONGSON_CFGX_L1_UNIFY	BIT(1)
#define LOONGSON_CFGX_L1_DPRE	BIT(2)
#define LOONGSON_CFGX_L2_IUPRE	BIT(3)
#define LOONGSON_CFGX_L2_IUUNIFY	BIT(4)
#define LOONGSON_CFGX_L2_IUPRIV	BIT(5)
#define LOONGSON_CFGX_L2_IUINCL	BIT(6)
#define LOONGSON_CFGX_L2_DPRE	BIT(7)
#define LOONGSON_CFGX_L2_DPRIV	BIT(8)
#define LOONGSON_CFGX_L2_DINCL	BIT(9)
#define LOONGSON_CFGX_L3_IUPRE	BIT(10)
#define LOONGSON_CFGX_L3_IUUNIFY	BIT(11)
#define LOONGSON_CFGX_L3_IUPRIV	BIT(12)
#define LOONGSON_CFGX_L3_IUINCL	BIT(13)
#define LOONGSON_CFGX_L3_DPRE	BIT(14)
#define LOONGSON_CFGX_L3_DPRIV	BIT(15)
#define LOONGSON_CFGX_L3_DINCL	BIT(16)

#define LOONGSON_CFGX1 		0x11
#define LOONGSON_CFGX1_WAYS_M	GENMASK(15, 0)
#define LOONGSON_CFGX1_SETS_M	GENMASK(23, 16)
#define LOONGSON_CFGX1_SIZE_M	GENMASK(30, 24)
#define LOONGSON_CFGX1_WAYS 0
#define LOONGSON_CFGX1_SETS	16
#define LOONGSON_CFGX1_SIZE 24

#define LOONGSON_CFGX2 		0x12
#define LOONGSON_CFGX2_WAYS_M	GENMASK(15, 0)
#define LOONGSON_CFGX2_SETS_M	GENMASK(23, 16)
#define LOONGSON_CFGX2_SIZE_M	GENMASK(30, 24)
#define LOONGSON_CFGX2_WAYS 0
#define LOONGSON_CFGX2_SETS	16
#define LOONGSON_CFGX2_SIZE 24

#define LOONGSON_CFGX3 		0x13
#define LOONGSON_CFGX3_WAYS_M	GENMASK(15, 0)
#define LOONGSON_CFGX3_SETS_M	GENMASK(23, 16)
#define LOONGSON_CFGX3_SIZE_M	GENMASK(30, 24)
#define LOONGSON_CFGX3_WAYS 0
#define LOONGSON_CFGX3_SETS	16
#define LOONGSON_CFGX3_SIZE 24

#define LOONGSON_CFGX4 		0x14
#define LOONGSON_CFGX4_WAYS_M	GENMASK(15, 0)
#define LOONGSON_CFGX4_SETS_M	GENMASK(23, 16)
#define LOONGSON_CFGX4_SIZE_M	GENMASK(30, 24)
#define LOONGSON_CFGX4_WAYS 0
#define LOONGSON_CFGX4_SETS	16
#define LOONGSON_CFGX4_SIZE 24

#define LOONGSON_CFG3X 		0x30
#define LOONGSON_CFG3X_MCSR_LCK	BIT(0)
#define LOONGSON_CFG3X_NAP_EN	BIT(1)
#define LOONGSON_CFG3X_VFPU_CG	BIT(2)
#define LOONGSON_CFG3X_RAM_CG	BIT(3)

static inline u32 csr_readl(u32 reg)
{
	return __lcsr_iocsrrd_w(reg);
}

static inline u64 csr_readq(u32 reg)
{
	return __lcsr_iocsrrd_d(reg);
}

static inline void csr_writel(u32 val, u32 reg)
{
	__lcsr_iocsrwr_w(val, reg);
}

static inline void csr_writeq(u64 val, u32 reg)
{
	__lcsr_iocsrwr_d(val, reg);
}
static inline unsigned long long read_guest_csr(int reg)
{
	unsigned long long __res;

	__asm__ __volatile__(
		"parse_r __res,%0\n\t"
		"parse_r reg,%1\n\t"
		".word (0xc8040118 | (reg << 21) | (__res << 11))\n\t"
		:"=r"(__res)
		:"r"(reg)
		:
		);
	return __res;
}

static inline void write_guest_csr(int reg,unsigned long long val)
{
	__asm__ __volatile__(
		"parse_r reg,%0\n\t"
		"parse_r val,%1\n\t"
		".word (0xc8050118 | (reg << 21) | (val << 11))\n\t"
		:
		:"r"(reg),"r"(val)
		:
		);
}

static inline unsigned long long dread_guest_csr(int reg)
{
	unsigned long long __res;

	__asm__ __volatile__(
		"parse_r __res,%0\n\t"
		"parse_r reg,%1\n\t"
		".word (0xc8060118 | (reg << 21) | (__res << 11))\n\t"
		:"=r"(__res)
		:"r"(reg)
		:
		);
	return __res;
}

static inline void dwrite_guest_csr(int reg,unsigned long long val)
{
	__asm__ __volatile__(
		"parse_r reg,%0\n\t"
		"parse_r val,%1\n\t"
		".word (0xc8070118 | (reg << 21) | (val << 11))\n\t"
		:
		:"r"(reg),"r"(val)
		:
		);
}

#define LOONGSON_CSR_FEATURES	0x8
#define LOONGSON_CSRF_TEMP	BIT_ULL(0)
#define LOONGSON_CSRF_NODECNT	BIT_ULL(1)
#define LOONGSON_CSRF_MSI	BIT_ULL(2)
#define LOONGSON_CSRF_EXTIOI	BIT_ULL(3)
#define LOONGSON_CSRF_IPI	BIT_ULL(4)
#define LOONGSON_CSRF_FREQ	BIT_ULL(5)
#define LOONGSON_CSRF_GMOD	BIT_ULL(9)

#define LOONGSON_CSR_VENDOR	0x10 /* Vendor name string, should be "Loongson" */
#define LOONGSON_CSR_CPUNAME	0x20 /* Processor name string */
#define LOONGSON_CSR_NODECNT	0x408
#define LOONGSON_CSR_OTHER_FUNC	0x420
#define CSR_OTHER_FUNC_EXT_INT_EN	BIT_ULL(48)
#define LS_CSR_ANYSEND_OTHER_FUNC	0x424
#define CSR_OTHER_FUNC_EXT_IOI	(CSR_OTHER_FUNC_EXT_INT_EN >> 32)
#define LOONGSON_CSR_CPUTEMP	0x428

/* PerCore CSR, only accessable by local cores */
#define LOONGSON_CSR_IPI_STATUS	0x1000
#define LOONGSON_CSR_IPI_EN	0x1004
#define LOONGSON_CSR_IPI_SET	0x1008
#define LOONGSON_CSR_IPI_CLEAR	0x100c
#define LOONGSON_CSR_MAIL_BUF0	0x1020
#define LOONGSON_CSR_MAIL_BUF1	0x1028
#define LOONGSON_CSR_MAIL_BUF2	0x1030
#define LOONGSON_CSR_MAIL_BUF3	0x1038
#define LOONGSON_CSR_IPI_SEND	0x1040

#define CSR_IPI_SEND_IP_SHIFT	0
#define CSR_IPI_SEND_CPU_SHIFT	16
#define CSR_IPI_SEND_BLOCKING	BIT(31)

#define LOONGSON_CSR_MAIL_SEND	0x1048

#define CSR_MAIL_SEND_BLOCKING	BIT_ULL(31)
#define CSR_MAIL_SEND_BOX_SHIFT	2
#define CSR_MAIL_SEND_BOX_LOW(box)	(box << 1)
#define CSR_MAIL_SEND_BOX_HIGH(box)	((box << 1) + 1)
#define CSR_MAIL_SEND_CPU_SHIFT	16
#define CSR_MAIL_SEND_BUF_SHIFT	32
#define CSR_MAIL_SEND_H32_MASK	0xFFFFFFFF00000000ULL

#define LOONGSON_CSR_ANY_SEND	0x1158

#define CSR_ANY_SEND_BLOCKING	BIT_ULL(31)
#define CSR_ANY_SEND_NODE_SHIFT	18
#define CSR_ANY_SEND_MASK_SHIFT	27
#define CSR_ANY_SEND_BUF_SHIFT	32
#define CSR_ANY_SEND_H32_MASK	0xFFFFFFFF00000000ULL

#define LOONGSON_CSR_EXTIOI_NODEMAP_BASE	0x14a0
#define LOONGSON_CSR_EXTIOI_IPMAP_BASE		0x14c0
#define LOONGSON_CSR_EXTIOI_EN_BASE		0x1600
#define CSR_EXTIOI_VECTOR_NUM			256
#define LOONGSON_CSR_EXTIOI_BOUNCE_BASE		0x1680
#define LOONGSON_CSR_EXTIOI_ISR_BASE		0x1800
#define LOONGSON_CSR_EXTIOI_ROUTE_BASE		0x1c00

static inline u64 drdtime(void)
{
	int rID = 0;
	u64 val = 0;

	__asm__ __volatile__(
		"rdtime.d %0, %1 \n\t"
		:"=r"(val),"=r"(rID)
		:
		);
	return val;
}

static inline void csr_any_send(unsigned int addr, unsigned int data, unsigned int data_mask, unsigned int node)
{
	uint64_t val = 0;

	val = CSR_ANY_SEND_BLOCKING | addr;
	val |= (node << CSR_ANY_SEND_NODE_SHIFT);
	val |= (data_mask << CSR_ANY_SEND_MASK_SHIFT);
	val |= ((uint64_t)data << CSR_ANY_SEND_BUF_SHIFT);
	__lcsr_iocsrwr_d(val, LOONGSON_CSR_ANY_SEND);
}

#endif /* !LOONGSON_CPU_H_ */
